package com.github.jcornaz.kwik.generator.api

import kotlin.random.Random

/**
 * Random data generator capable of producing test data for property testing.
 */
interface Generator<T> {

    /**
     * Samples of values that should always be tested
     *
     * @deprecated This function will be removed soon. Instead, the samples will have higher probability to be generated by [generate].
     *             In the mean time favor usage of `withSamples` or `frequency`.
     */
    @Deprecated("Use withSamples instead")
    val samples: Set<T> get() = emptySet()

    /**
     * Returns a sequence of random value.
     *
     * **Must be pure**: It must always return the same value sequence for the same given [seed]
     *
     * @param seed Seed of the random generation.
     *             Two invocations of the function with the same seed must return the same value sequence
     */
    @Deprecated("Use randomSequence extension function instead", ReplaceWith("randomSequence(seed)"))
    fun randoms(seed: Long): Sequence<T> = randomSequence(seed)

    /**
     * Returns a random value using the given [random].
     *
     * This function can mutate [random]. But it should always return the same value for a given [Random] state.
     * Ex. `generate(Random(0L))` should always return the same value.
     */
    fun generate(random: Random): T

    companion object {

        /**
         * Create a simple random [Generator].
         *
         * @param next Function that will be invoked to get a new random parameter.
         *             The function should use the given [Random] generator to ensure predictability of the values
         */
        fun <T> create(next: (Random) -> T): Generator<T> = object : Generator<T> {
            override val samples: Set<T> get() = emptySet()

            override fun generate(random: Random): T = next(random)
        }

        /**
         * Create a random [Generator] generating values out of the given [samples]
         */
        fun <T> of(vararg samples: T): Generator<T> {
            require(samples.isNotEmpty()) { "No given sample" }

            return create { samples.random(it) }
        }
    }
}

/**
 * Returns a sequence of random values.
 *
 * @param seed Seed of the random generation.
 *             Two invocations of the function with the same seed must return the same value sequence
 */
fun <T> Generator<T>.randomSequence(seed: Long): Sequence<T> =
    randomSequence(seed) { generate(it) }

/**
 * Returns a sequence of random values.
 *
 * @param seed Seed of the random generation.
 *             Two invocations of the function with the same seed will return the same value sequence
 * @param next Function that will be invoked to get a new random parameter.
 *             The function should use the given [Random] generator to ensure predictability of the values
 */
fun <T> randomSequence(seed: Long, next: (Random) -> T): Sequence<T> = sequence {
    val rng = Random(seed)

    while (true) yield(next(rng))
}
